"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const http = require("http");
const utils = require("../utils");
const verify = require("../verify");
const homedir = require('os').homedir();
const uuidv4 = require('uuid/v4');
const rp = require("../request");
/**
 * Describes a way to subscribe to an event.
 */
class Subscribe {
    /**
     *
     * @param {Microservice} microservice The given {@link Microservice}
     * @param {Object} _arguments The given arguments
     * @param {Run} run The {@link Run} object that started the event
     */
    constructor(microservice, _arguments, run) {
        this.microservice = microservice;
        this._arguments = _arguments;
        this.run = run;
    }
    /**
     * Subscribes you to the given event.
     *
     * @param {String} action The given action
     * @param {String} event The given event
     * @param  {string} ip The given IP address
     */
    go(action, event) {
        return __awaiter(this, void 0, void 0, function* () {
            this.omgJson = JSON.parse(fs.readFileSync(`${homedir}/.omg.json`, 'utf8'));
            this.action = this.microservice.getAction(action);
            this.event = this.action.getEvent(event);
            if (!this.event.areRequiredArgumentsSupplied(this._arguments)) {
                throw `Failed subscribing to event: \`${event}\`. Need to supply required arguments: \`${this.event.requiredArguments.toString()}\``;
            }
            verify.verifyArgumentTypes(this.event, this._arguments);
            this.castTypes();
            verify.verifyArgumentConstrains(this.event, this._arguments);
            const server = this.startOMGServer();
            const port = yield utils.getOpenPort();
            server.listen({ port, hostname: '127.0.0.1' });
            this.id = uuidv4();
            yield this.subscribe(port);
        });
    }
    /**
     * Subscribes to an event, the try catch is needed for the slow start of the container. (same issue with http commands)
     *
     * @param {Number} port The given port to request on
     * @param  {string} ip The given IP address
     * @return {Promise<void>}
     */
    subscribe(port) {
        return __awaiter(this, void 0, void 0, function* () {
            yield rp.makeRequest({
                method: this.event.subscribe.method,
                uri: `http://localhost:${this.omgJson[process.cwd()].ports[this.event.subscribe.port]}${this.event.subscribe.path}`,
                body: {
                    id: this.id,
                    endpoint: `http://host.docker.internal:${port}`,
                    event: this.event.name,
                    data: this._arguments
                },
                json: true
            });
        });
    }
    /**
     * Starts a server for a streaming service to POST back to.
     *
     * @return {Server} The server
     */
    startOMGServer() {
        return http.createServer((req, res) => {
            if (req.method === 'POST') {
                req.on('data', (data) => __awaiter(this, void 0, void 0, function* () {
                    try {
                        verify.verifyOutputType(this.event, data.toString());
                        if (this.event.output &&
                            this.event.output.type &&
                            (this.event.output.type === 'map' ||
                                this.event.output.type === 'object')) {
                            utils.log(JSON.stringify(JSON.parse(data), null, 2));
                        }
                        else {
                            utils.log(data);
                        }
                    }
                    catch (e) {
                        utils.error(e);
                    }
                    res.end('Done');
                }));
            }
        });
    }
    /**
     * Cast the types of the arguments. Everything comes in as a string so it's important to convert to given type.
     */
    castTypes() {
        const argumentList = Object.keys(this._arguments);
        for (let i = 0; i < argumentList.length; i += 1) {
            const argument = this.event.getArgument(argumentList[i]);
            this._arguments[argument.name] = utils.typeCast[argument.type](this._arguments[argument.name]);
        }
    }
    /**
     * Unsubscribe this {@link Subscribe}'s {@link Event}.
     */
    unsubscribe() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.event.unsubscribe === null) {
                return;
            }
            yield rp.makeRequest({
                method: this.event.unsubscribe.method,
                uri: `http://localhost:${this.omgJson[process.cwd()].ports[this.event.unsubscribe.port]}${this.event.unsubscribe.path}`,
                body: {
                    id: this.id
                },
                json: true
            });
        });
    }
}
exports.default = Subscribe;
//# sourceMappingURL=Subscribe.js.map