"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Run_1 = require("./Run");
const utils = require("../../utils");
const verify = require("../../verify");
const fs = require('fs');
/**
 * Represents a docker exec execution of an {@link Action}.
 */
class FormatRun extends Run_1.default {
    /**
     * Builds a {@link FormatRun}.
     *
     * @param {String} dockerImage The given docker image
     * @param {Microservice} microservice The given {@link Microservice}
     * @param {Object} _arguments The given argument map
     * @param {Object} environmentVariables the given environment  map
     */
    constructor(dockerImage, microservice, _arguments, environmentVariables) {
        super(dockerImage, microservice, _arguments, environmentVariables);
    }
    /** @inheritdoc */
    exec(action) {
        return __awaiter(this, void 0, void 0, function* () {
            this.action = this.microservice.getAction(action);
            this.preChecks();
            this.verification();
            const output = yield this.runDockerExecCommand(this.containerID);
            verify.verifyOutputType(this.action, output);
            if (this.action.output &&
                this.action.output.type &&
                (this.action.output.type === 'map' ||
                    this.action.output.type === 'object')) {
                if (this.action.output.properties) {
                    verify.verifyProperties(this.action, output);
                }
                return JSON.stringify(JSON.parse(output.trim()), null, 2);
            }
            return output.trim();
        });
    }
    /**
     * Starts the docker container based of this {@link Exec}'s {@link Microservice}'s {@link Lifecycle}. If null,
     * the container will be started with the command: `tail -f /dev/null``.
     *
     * @return {Promise<String>} The id of the started container
     */
    startService() {
        return __awaiter(this, void 0, void 0, function* () {
            this.setDefaultEnvironmentVariables();
            const lifecycle = this.microservice.lifecycle;
            if (lifecycle !== null && lifecycle.startup !== null) {
                const container = yield utils.docker.createContainer({
                    Image: this.dockerImage,
                    Cmd: lifecycle.startup,
                    Env: this.formatEnvironmentVariables()
                });
                yield container.start();
                this.containerID = container.$subject.id;
            }
            else {
                const container = yield utils.docker.createContainer({
                    Image: this.dockerImage,
                    Cmd: ['tail', '-f', '/dev/null'],
                    Env: this.formatEnvironmentVariables()
                });
                yield container.start();
                this.containerID = container.$subject.id;
            }
            return this.containerID;
        });
    }
    /**
     * Runs a given command via Docker API.
     *
     * @param {String} containerID The given id of the docker container
     * @return {Promise<String>} stdout if command runs with exit code 0, otherwise stderror
     */
    runDockerExecCommand(containerID) {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            const cmd = this.action.format.command;
            cmd.push(this.formatExec());
            const exec = yield container.exec({
                Cmd: cmd,
                AttachStdin: true,
                AttachStdout: true,
                Tty: true
            });
            const data = yield new Promise((resolve, reject) => {
                exec.start({ stdin: true, stdout: true }, (err, stream) => {
                    const data = [];
                    if (err) {
                        throw err;
                    }
                    else {
                        stream.on('data', chunk => {
                            data.push(chunk);
                        });
                        stream.on('end', () => {
                            resolve(Buffer.concat(data)
                                .toString()
                                .trim()
                                .substring(8));
                        });
                    }
                });
            });
            if ((yield exec.inspect()).ExitCode !== 0) {
                throw data;
            }
            return data;
        });
    }
    /**
     * Formats this {@link Microservice}'s {@link Argument}s into stringified JSON.
     *
     * @return {String} The JSON string
     * @private
     */
    formatExec() {
        if (this.action.arguments.length > 0) {
            return JSON.stringify(this._arguments);
        }
        return '';
    }
    /**
     * @param  {any} args
     */
    setArgs(args) { }
}
exports.default = FormatRun;
//# sourceMappingURL=FormatRun.js.map