"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const rp = require("request-promise");
const querystring = require("querystring");
const Run_1 = require("./Run");
const verify = require("../../verify");
const FormData = require("form-data");
/**
 * Represents a http execution of an {@link Action}.
 */
class HttpRun extends Run_1.default {
    /**
     * Builds a {@link HttpRun}.
     *
     * @param {String} dockerImage The given docker image
     * @param {Microservice} microservice The given {@link Microservice}
     * @param {Object} _arguments The given argument map
     * @param {Object} environmentVariables the given environment  map
     */
    constructor(dockerImage, microservice, _arguments, environmentVariables) {
        super(dockerImage, microservice, _arguments, environmentVariables);
    }
    /** @inheritdoc */
    exec(action, tmpRetryExec = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.action = this.microservice.getAction(action);
            this.preChecks();
            this.verification();
            const output = yield this.httpCommand(this.portMap[this.action.http.port], tmpRetryExec);
            if (tmpRetryExec) {
                verify.verifyOutputType(this.action, output.body.trim());
                return output;
            }
            else {
                verify.verifyOutputType(this.action, output.trim());
                if (this.action.output &&
                    this.action.output.type &&
                    (this.action.output.type === 'map' ||
                        this.action.output.type === 'object')) {
                    if (this.action.output.properties) {
                        verify.verifyProperties(this.action, output);
                    }
                    return JSON.stringify(JSON.parse(output.trim()), null, 2);
                }
                return output.trim();
            }
        });
    }
    /**
     * Run this {@link HttpRun}'s {@link Action} that interfaces via HTTP.
     *
     * @param {Number} port The given sever started in Docker
     * @param {boolean} tmpRetryExec Temporary boolean
     * @return {Promise<String>} The response of the Http request
     */
    httpCommand(port, tmpRetryExec = false) {
        return __awaiter(this, void 0, void 0, function* () {
            // Temporary, remove when health is mandatory (put <string> back too)
            const httpData = this.formatHttp(port);
            const opts = {
                method: this.action.http.method.toUpperCase(),
                resolveWithFullResponse: tmpRetryExec,
                uri: httpData.uri
            };
            if (this.action.http.method === 'post' ||
                this.action.http.method === 'put') {
                switch (this.action.http.contentType) {
                    case 'multipart/form-data': {
                        opts.formData = this.jsonToFormData(httpData.jsonData);
                        opts.headers = opts.formData.getHeaders();
                        break;
                    }
                    case 'application/x-www-form-urlencoded':
                        opts.body = Object.keys(httpData.jsonData)
                            .map(key => {
                            return (encodeURIComponent(key) +
                                '=' +
                                encodeURIComponent(httpData.jsonData[key]));
                        })
                            .join('&');
                        opts.headers = {
                            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
                        };
                        break;
                    case 'application/json':
                    default: {
                        opts.body = JSON.stringify(httpData.jsonData);
                        opts.headers = {
                            'Content-Type': 'application/json'
                        };
                    }
                }
            }
            /**
             * Since rp(opts) doesn't pass the tests, I use this
             * in order to add resolveWithFullResponse
             *
             * Since rp.get/post/put/delete cannot use
             * resolveWithFullResponse, I must keep the switchCase for now
             */
            if (tmpRetryExec) {
                return yield rp(opts);
            }
            let data = {};
            switch (this.action.http.method) {
                case 'get':
                    data = yield rp.get(opts.uri);
                    break;
                case 'post': {
                    data = yield rp.post(opts.uri, {
                        headers: opts.headers,
                        body: opts.body
                    });
                    break;
                }
                case 'put':
                    data = yield rp.put(opts.uri, {
                        headers: opts.headers,
                        body: opts.body
                    });
                    break;
                case 'delete':
                    data = yield rp.delete(opts.uri);
                    break;
            }
            return data;
        });
    }
    /**
     * Formats an Http request based on this {@link HttpRun}'s {@link Action}.
     *
     * @param {Number} port The given server info
     * @return {{url: String, jsonData: Object}} The url and data
     */
    formatHttp(port) {
        const jsonData = {};
        const queryParams = {};
        let uri = `http://localhost:${port}${this.action.http.path}`;
        for (let i = 0; i < this.action.arguments.length; i += 1) {
            const argument = this.action.arguments[i];
            switch (this.action.getArgument(argument.name).in) {
                case 'query':
                    if (!_.isUndefined(this._arguments[argument.name])) {
                        queryParams[argument.name] = this._arguments[argument.name];
                    }
                    break;
                case 'path':
                    uri = uri.replace(`{${argument.name}}`, this._arguments[argument.name]);
                    break;
                case 'requestBody':
                    jsonData[argument.name] = this._arguments[argument.name];
                    break;
            }
        }
        if (querystring.stringify(queryParams) !== '') {
            uri = `${uri}?${querystring.stringify(queryParams)}`;
        }
        return {
            uri,
            jsonData
        };
    }
    /**
     * Makes a FormData out of a JSON object
     * Not used yet
     *
     * @param  {any} json JSON input
     * @return {FormData} FormData
     */
    jsonToFormData(json) {
        const form = new FormData();
        Object.keys(json).forEach(key => {
            form.append(key, json[key]);
        });
        return form;
    }
    /**
     * @param  {any} args
     */
    setArgs(args) { }
}
exports.default = HttpRun;
//# sourceMappingURL=HttpRun.js.map