"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils = require("../../utils");
const verify = require("../../verify");
const rp = require("request-promise");
const $ = require("shelljs");
/**
 * Used to represent a way to execute a {@link Microservice}'s {@link Action}s.
 */
class Run {
    /**
     * Use to help build a {@link FormatRun}, {@link HttpRun}, or an {@link EventRun}.
     *
     * @param {String} dockerImage The given docker image
     * @param {Microservice} microservice The given {@link Microservice}
     * @param {Object} _arguments The argument map
     * @param {Object} environmentVariables The environment map
     */
    constructor(dockerImage, microservice, _arguments, environmentVariables) {
        this.portMap = {};
        this.exposedPorts = {};
        this.portBindings = {};
        this.containerID = null;
        this.dockerImage = dockerImage;
        this.microservice = microservice;
        this._arguments = _arguments;
        this.environmentVariables = environmentVariables;
        this.dockerServiceId = null;
        this.action = null;
    }
    /**
     * Checks if required arguments and environment variables are given and will also set their default values.
     */
    preChecks() {
        this.setDefaultArguments();
        if (!this.action.areRequiredArgumentsSupplied(this._arguments)) {
            throw `Need to supply required arguments: \`${this.action.requiredArguments.toString()}\``;
        }
        if (!this.microservice.areRequiredEnvironmentVariablesSupplied(this.environmentVariables)) {
            throw `Need to supply required environment variables: \`${this.microservice.requiredEnvironmentVariables.toString()}\``;
        }
    }
    /**
     * Runs verification on arguments and environment variables.
     */
    verification() {
        verify.verifyArgumentTypes(this.action, this._arguments);
        this.castTypes();
        verify.verifyArgumentConstrains(this.action, this._arguments);
        verify.verifyEnvironmentVariableTypes(this.microservice, this.environmentVariables);
        verify.verifyEnvironmentVariablePattern(this.microservice, this.environmentVariables);
    }
    /**
     * Sets a {@link Action}'s default arguments.
     */
    setDefaultArguments() {
        for (let i = 0; i < this.action.arguments.length; i += 1) {
            const argument = this.action.arguments[i];
            if (!this._arguments[argument.name]) {
                if (argument.default !== null) {
                    if (typeof argument.default === 'object') {
                        this._arguments[argument.name] = JSON.stringify(argument.default);
                    }
                    else {
                        this._arguments[argument.name] = argument.default + '';
                    }
                }
            }
        }
    }
    /**
     * Sets a {@link Microservice}'s default {@link EnvironmentVariable}s and variables from the system environment variables.
     *
     * @param {boolean} [inheritEnv=false] Boolean that allows to get env from host env or not
     */
    setDefaultEnvironmentVariables(inheritEnv) {
        for (let i = 0; i < this.microservice.environmentVariables.length; i += 1) {
            const environmentVariable = this.microservice.environmentVariables[i];
            if (!this.environmentVariables[environmentVariable.name]) {
                if (environmentVariable.default !== null) {
                    this.environmentVariables[environmentVariable.name] =
                        environmentVariable.default;
                }
            }
        }
        if (inheritEnv) {
            for (let i = 0; i < this.microservice.environmentVariables.length; i++) {
                const environmentVariable = this.microservice.environmentVariables[i];
                if (process.env[environmentVariable.name]) {
                    this.environmentVariables[environmentVariable.name] =
                        process.env[environmentVariable.name];
                }
            }
        }
    }
    /**
     * Cast the types of the arguments. Everything comes in as a string so it's important to convert to given type.
     */
    castTypes() {
        const argumentList = Object.keys(this._arguments);
        for (let i = 0; i < argumentList.length; i++) {
            const argument = this.action.getArgument(argumentList[i]);
            this._arguments[argument.name] = utils.typeCast[argument.type](this._arguments[argument.name]);
        }
    }
    /**
     * Formats an object of environment variables to a `-e KEY='val'` style.
     *
     * @return {String} The formatted string
     */
    formatEnvironmentVariables() {
        const result = [];
        const keys = Object.keys(this.environmentVariables);
        for (let i = 0; i < keys.length; i += 1) {
            result.push(`${keys[i]}=${this.environmentVariables[keys[i]]}`);
        }
        return result;
    }
    /**
     * Checks it a Docker process is running.
     *
     * @return {Boolean} True if a Docker process is running, otherwise false
     */
    isDockerProcessRunning() {
        return false;
    }
    /**
     * @return {Promise} Return the host IP
     */
    static getHostIp() {
        const hostDomain = 'host.docker.internal';
        return new Promise((resolve, reject) => {
            $.exec('ip -4 addr show docker0', { silent: true }, (code, stdout, stderr) => {
                if (code === 0) {
                    resolve(`${hostDomain}:${stdout.match(/inet ([\d.]+)/)[1]}`);
                }
                reject();
            });
        });
    }
    /**
     * Starts the server for the HTTP command based off the lifecycle provided in the microservice.yml and builds port mapping.
     *
     * @param {boolean} [inheritEnv=false] Boolean that allows to get env from host env or not
     * @return {String} The id of the container
     */
    startService(inheritEnv = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.setDefaultEnvironmentVariables(inheritEnv);
            const neededPorts = utils.getNeededPorts(this.microservice);
            const openPorts = [];
            while (neededPorts.length !== openPorts.length) {
                const possiblePort = yield utils.getOpenPort();
                if (!openPorts.includes(possiblePort)) {
                    openPorts.push(possiblePort);
                }
            }
            for (let i = 0; i < neededPorts.length; i += 1) {
                this.portMap[neededPorts[i]] = openPorts[i];
                this.exposedPorts[`${neededPorts[i]}/tcp`] = {};
                this.portBindings[`${neededPorts[i]}/tcp`] = [
                    { HostPort: openPorts[i].toString() }
                ];
            }
            const container = yield utils.docker.createContainer({
                Image: this.dockerImage,
                Cmd: this.microservice.lifecycle
                    ? this.microservice.lifecycle.startup
                    : null,
                Env: this.formatEnvironmentVariables(),
                ExposedPorts: this.exposedPorts,
                HostConfig: {
                    PortBindings: this.portBindings,
                    ExtraHosts: !['darwin', 'win32'].includes(process.platform) ? [yield Run.getHostIp()] : []
                }
            });
            yield container.start();
            this.containerID = container.$subject.id;
            return this.containerID;
        });
    }
    /**
     * Stops a running Docker service, if it is already stopped just return the id of that container.
     *
     * @return {String} The containerID that has been stopped
     */
    stopService() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            yield container.kill();
            return this.containerID;
        });
    }
    /**
     * Removes a stopped Docker service
     *
     * @return {String} The containerID that has been stopped
     */
    removeContainer() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            yield container.remove();
            return this.containerID;
        });
    }
    /**
     * Checks if there is a Docker process running.
     *
     * @return {Boolean} True if running, otherwise false
     */
    isRunning() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            return (yield container.inspect()).State.Running;
        });
    }
    /**
     * Returns docker inspect result as a Promise
     *
     * @return {Promise}
     */
    getInspect() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            return yield container.inspect();
        });
    }
    /**
     * Returns docker stats result as a Promise
     *
     * @return {Promise}
     */
    getStats() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            return yield container.stats({ stream: false });
        });
    }
    /**
     * Gets the Docker logs.
     *
     * @return {String} The Docker logs
     */
    getStderr() {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            return (yield container.logs({ stderr: true })).toString().trim();
        });
    }
    /**
     * Returns docker logs result as a Promise
     *
     * @param  {number} since Optional parameters filtering logs 'since' UNIX timestamp
     * @return {Promise}
     */
    getLogs(since) {
        return __awaiter(this, void 0, void 0, function* () {
            const container = utils.docker.getContainer(this.containerID);
            const logs = yield container.logs({
                stderr: true,
                stdout: true,
                since: since !== null ? since : 0
            });
            return logs.toString().trim();
        });
    }
    /**
     * Gets and return bond port
     *
     * @return {any} Bond ports
     */
    getPortbindings() {
        return this.portBindings;
    }
    /**
     * Gets and return bond port only for forwards
     * @return {any} Bond ports
     */
    get forwardPortsBindings() {
        const bindings = {};
        const ports = utils.getForwardPorts(this.microservice);
        for (const forward of this.microservice.forwards) {
            bindings[forward.name] = {
                host: this.portBindings[`${forward.port}/tcp`],
                container: forward.port
            };
        }
        return bindings;
    }
    /**
     * Gets health status of the service
     *
     * @param  {number} boundPort port used for health checking
     * @param  {number} timeout Optionnal timeout, used during healthcheck
     * @return {Promise} Empty promise
     */
    isHealthy(boundPort, timeout) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                const promise = rp.get({
                    uri: `http://localhost:${boundPort}${this.microservice.health.path}`,
                    method: 'GET',
                    resolveWithFullResponse: true
                });
                if (timeout) {
                    setTimeout(() => {
                        reject(promise.cancel());
                    }, timeout);
                }
                promise
                    .then(response => {
                    switch (response.statusCode / 100) {
                        case 2:
                        case 3:
                            resolve();
                            break;
                        default:
                            reject();
                            break;
                    }
                })
                    .catch(e => {
                    reject();
                });
            });
        });
    }
    /**
     * Does health check for the service until it's healthy or unhealthy
     *
     * @return {Promise} health status as a boolean
     */
    healthCheck() {
        return __awaiter(this, void 0, void 0, function* () {
            const timeout = 500;
            const interval = 100;
            const retries = 100;
            let boundPort = -1;
            Object.keys(this.portBindings).forEach(p => {
                const port = parseInt(p.match(/[\d]*/)[0], 10);
                if (port === utils.getHealthPort(this.microservice)) {
                    boundPort = this.portBindings[p][0].HostPort;
                }
            });
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                yield utils.sleep(10);
                for (let i = retries; i > 0; i--) {
                    if (this.microservice.health) {
                        yield this.isHealthy(boundPort, timeout)
                            .then(() => {
                            i = 0;
                            resolve(true);
                        })
                            .catch(() => __awaiter(this, void 0, void 0, function* () {
                            yield utils.sleep(interval);
                        }));
                    }
                }
                reject(false);
            }));
        });
    }
}
exports.default = Run;
//# sourceMappingURL=Run.js.map