"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Run_1 = require("./Run");
const verify_1 = require("../../verify");
const rp = require("request-promise");
const _ = require("underscore");
const querystring = require("querystring");
const fs = require("fs");
const utils = require("../../utils");
const verify = require("../../verify");
const homedir = require('os').homedir();
/**
 * Used to represent a way to execute an {@link Action} in the UI.
 */
class UIRun extends Run_1.default {
    /**
     * Constructor
     *
     * @param  {string} dockerImage
     * @param  {Microservice} microservice
     * @param  {any} envs
     */
    constructor(dockerImage, microservice, envs) {
        super(dockerImage, microservice, null, envs);
    }
    /**
     * Execs the given action
     *
     * @param  {string} action
     * @return {Promise}
     */
    exec(action) {
        return __awaiter(this, void 0, void 0, function* () {
            this.action = this.microservice.getAction(action);
            this.preChecks();
            try {
                this.verification();
            }
            catch (e) {
                utils.error(e);
            }
            if (this.action.http === null) {
                this.omgJsonFileHandler();
                return '';
            }
            else {
                const output = yield this.httpCommand(this.portMap[this.action.http.port]);
                verify_1.verifyOutputType(this.action, output.trim());
                if (this.action.output &&
                    this.action.output.type &&
                    (this.action.output.type === 'map' ||
                        this.action.output.type === 'object')) {
                    if (this.action.output.properties) {
                        verify.verifyProperties(this.action, output);
                    }
                    return JSON.stringify(JSON.parse(output.trim()), null, 2);
                }
                return output.trim();
            }
        });
    }
    /**
     * Handle the `.omg.json` state file.
     */
    omgJsonFileHandler() {
        let data = {};
        if (fs.existsSync(`${homedir}/.omg.json`)) {
            data = JSON.parse(fs.readFileSync(`${homedir}/.omg.json`).toString());
        }
        data[process.cwd()] = {
            container_id: this.containerID,
            ports: {}
        };
        const neededPorts = Object.keys(this.portMap);
        for (let i = 0; i < neededPorts.length; i += 1) {
            data[process.cwd()].ports[neededPorts[i]] = this.portMap[neededPorts[i]];
        }
        fs.writeFileSync(`${homedir}/.omg.json`, JSON.stringify(data), 'utf8');
    }
    /** @inheritdoc */
    isDockerProcessRunning() {
        return this.containerID !== null;
    }
    /**
     * Run this {@link HttpRun}'s {@link Action} that interfaces via HTTP.
     *
     * @param {Number} port The given sever started in Docker
     * @return {Promise<String>} The response of the Http request
     */
    httpCommand(port) {
        return __awaiter(this, void 0, void 0, function* () {
            let data;
            const httpData = this.formatHttp(port);
            switch (this.action.http.method) {
                case 'get':
                    data = yield rp.get(httpData.url);
                    break;
                case 'post':
                    data = yield rp.post(httpData.url, {
                        headers: {
                            'Content-Type': 'application/json'
                        },
                        body: JSON.stringify(httpData.jsonData)
                    });
                    break;
                case 'put':
                    data = yield rp.put(httpData.url, {
                        headers: {
                            'Content-Type': 'application/json'
                        },
                        body: JSON.stringify(httpData.jsonData)
                    });
                    break;
                case 'delete':
                    data = yield rp.delete(httpData.url);
                    break;
            }
            return data;
        });
    }
    /**
     * Formats an Http request based on this {@link HttpRun}'s {@link Action}.
     *
     * @param {Number} port The given server info
     * @return {{url: String, jsonData: Object}} The url and data
     */
    formatHttp(port) {
        const jsonData = {};
        const queryParams = {};
        let url = `http://localhost:${port}${this.action.http.path}`;
        for (let i = 0; i < this.action.arguments.length; i += 1) {
            const argument = this.action.arguments[i];
            switch (this.action.getArgument(argument.name).in) {
                case 'query':
                    if (!_.isUndefined(this._arguments[argument.name])) {
                        queryParams[argument.name] = this._arguments[argument.name];
                    }
                    break;
                case 'path':
                    url = url.replace(`{${argument.name}}`, this._arguments[argument.name]);
                    break;
                case 'requestBody':
                    jsonData[argument.name] = this._arguments[argument.name];
                    break;
            }
        }
        if (querystring.stringify(queryParams) !== '') {
            url = `${url}?${querystring.stringify(queryParams)}`;
        }
        return {
            url,
            jsonData
        };
    }
    /**
     * Sets the given args in to the container
     *
     * @param  {any} args
     */
    setArgs(args) {
        if (!args) {
            return;
        }
        this._arguments = args;
    }
}
exports.default = UIRun;
//# sourceMappingURL=UIRun.js.map