"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils = require("../../utils");
const http_1 = require("./wrappers/http");
const socket_io_1 = require("./wrappers/socket-io");
const omg_ui_1 = require("omg-ui");
const Subscribe_1 = require("../Subscribe");
const omg_validate_1 = require("omg-validate");
const Build_1 = require("../Build");
const Cli_1 = require("../../cli/Cli");
const RunFactory_1 = require("../run/RunFactory");
const fs = require("fs");
const open_1 = require("./wrappers/open");
const dockerode_1 = require("./wrappers/dockerode");
const LineUp = require('lineup');
const lineup = new LineUp();
/**
 * UIServer definition
 */
class UIServer {
    /**
     * Constructor
     *
     * @param  {any} options
     * @param  {any} microservice
     */
    constructor(options, microservice) {
        this.port = options.port;
        this.app = omg_ui_1.app();
        this.http = new http_1.default.Server(this.app);
        this.io = socket_io_1.default.listen(this.http);
        this.socket = null;
        this.microserviceStr = microservice;
        this.rebuildToggle = true;
        this.isClientConnected = false;
        this.inheritEnv = options.inheritEnv ? options.inheritEnv : false;
    }
    /**
     * Starts the UI server
     * @param {Boolean} [doOpen=false]
     */
    startUI(doOpen = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.io.on('connection', socket => {
                if (!this.isClientConnected) {
                    this.isClientConnected = true;
                    socket.removeAllListeners();
                    this.socket = socket;
                    this.sendFile(utils.getMicroserviceFilePath());
                    this.initListeners();
                    utils.log('Web client connected to socket.');
                }
                this.socket.on('disconnect', () => {
                    this.isClientConnected = false;
                    utils.log('Web client disconnected from socket.');
                });
            });
            if (!this.port) {
                this.port = yield utils.getOpenPort();
            }
            this.http.listen(this.port, () => __awaiter(this, void 0, void 0, function* () {
                lineup.sticker.note('');
                lineup.sticker.note(`${lineup.colors.yellow('OMG UI is currently a work-in-progress product.')}`);
                lineup.sticker.note(`${lineup.colors.yellow('It can have some issues, which you can report here:')}`);
                lineup.sticker.note(`${lineup.colors.yellow('https://github.com/microservices/omg/issues')}`);
                lineup.sticker.note('');
                lineup.sticker.show({
                    align: 'center',
                    color: 'red'
                });
                utils.log(`OMG UI started on http://localhost:${this.port}`);
                if (doOpen) {
                    yield open_1.default(`http://localhost:${this.port}`);
                }
            }));
        });
    }
    /**
     * Inits all socket listeners
     */
    initListeners() {
        this.socket.on('build', (data) => {
            this.buildImage(data);
        });
        this.socket.on('run', (data) => {
            this.runAction(data);
        });
        this.socket.on('health-check', () => {
            this.healthCheck();
        });
        this.socket.on('inspect', () => {
            this.dockerInspect();
        });
        this.socket.on('subscribe', (data) => {
            this.subscribeEvent(data);
        });
        this.socket.on('dockerLogs', () => {
            this.dockerLogs();
        });
        this.socket.on('start', (data) => {
            this.startContainer(data);
        });
        this.socket.on('stop', () => {
            this.stopContainer();
        });
        this.socket.on('container-stats', () => {
            this.dockerStats();
        });
        this.socket.on('rebuild', data => {
            this.rebuild(data);
        });
        this.socket.on('rebuild-toggle', data => {
            this.rebuildToggle = data;
        });
        this.socket.on('clear-container-logs', timestamp => {
            this.clogsSince = timestamp;
        });
        this.socket.on('microservice.yml', (data) => {
            const content = new Uint8Array(Buffer.from(data));
            fs.writeFile(utils.getMicroserviceFilePath(), content, err => {
                if (err)
                    throw err;
                this.validate();
                utils.log('microservice.yml file has been saved!');
            });
        });
    }
    /**
     * Rebuild image and restarts container
     *
     * @param  {any} data Data used to build image
     * @param  {string} microservice
     * @param  {Boolean} [bak=false]
     * @param  {string} appPath
     */
    rebuild(data, microservice, bak = false, appPath) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.rebuildToggle) {
                if (appPath) {
                    utils.log(`${appPath.substr(appPath.lastIndexOf('/') + 1)} changed. Rebuilding.`);
                }
                if (microservice) {
                    this.microserviceStr = microservice;
                }
                this.sendFile(utils.getMicroserviceFilePath());
                yield this.stopContainer();
                yield this.removeContainer();
                if (bak && this.rebuildBak) {
                    yield this.buildImage(this.rebuildBak.build);
                }
                else {
                    yield this.buildImage(data.build);
                    this.rebuildBak = { build: data.build, start: data.start };
                }
            }
        });
    }
    /**
     * Sends provided file raw content
     *
     * @param  {string} file
     */
    sendFile(file) {
        fs.readFile(file, 'utf8', (err, data) => {
            if (err)
                throw err;
            this.socket.emit('microservice.yml', data);
            this.validate();
        });
    }
    /**
     * socket.emit serializer
     *
     * @param  {string} room
     * @param  {ISocketNotif} msg
     */
    emit(room, msg) {
        this.socket.emit(room, msg);
    }
    /**
     * Removes socket listeners
     *
     * @return {Promise} a promise when listeners have been removed
     */
    removeListeners() {
        return this.socket.removeAllListeners();
    }
    /**
     * Wraps omg-cli validate command
     */
    validate() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (this.microserviceStr === 'ERROR_PARSING') {
                    this.emit('validate', {
                        notif: this.microserviceStr,
                        status: false
                    });
                }
                else {
                    utils.checkActionInterface(this.microserviceStr);
                    this.microservice = new omg_validate_1.Microservice(this.microserviceStr);
                    this.emit('validate', {
                        notif: JSON.stringify(this.microservice.rawData, null, 2),
                        status: true
                    });
                }
            }
            catch (e) {
                this.emit('validate', {
                    notif: JSON.stringify(e, null, 2),
                    status: false
                });
            }
            const owner = yield utils.createImageName(true);
            this.socket.emit('owner', {
                notif: typeof owner && owner.generated ? owner.owner : owner,
                generated: typeof owner && owner.generated,
                status: true
            });
        });
    }
    /**
     * Wraps omg-cli build command
     *
     * @param  {IDataBuild} data image name
     */
    buildImage(data) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Cli_1.default.checkDocker();
            this.emit('build', { notif: 'Building Docker image', status: true });
            try {
                const stream = yield new Build_1.default(data.name || (yield utils.createImageName())).go(false, true);
                const dockerode = new dockerode_1.default();
                const log = yield new Promise((resolve, reject) => {
                    dockerode.modem.followProgress(stream, (err, res) => err ? reject(err) : resolve(res));
                });
                for (const line in log) {
                    if (log[line].stream) {
                        this.socket.emit('build', {
                            status: true,
                            notif: log[line].stream.trim(),
                            build: true
                        });
                    }
                }
                this.socket.emit('build', {
                    status: true,
                    built: true
                });
                return data.name;
            }
            catch (e) {
                this.emit('build', {
                    status: false,
                    notif: `Failed to build: ${e}`
                });
            }
        });
    }
    /**
     * Inspects docker container then the sends result through socket
     */
    dockerInspect() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.dockerContainer) {
                const output = yield this.dockerContainer.getInspect();
                this.emit('inspect', {
                    notif: 'Docker inspected',
                    status: true,
                    log: output
                });
            }
            else {
                this.emit('inspect', { status: false, notif: 'Container not running' });
            }
        });
    }
    /**
     * Gets and returns docker container stats
     */
    dockerStats() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.dockerContainer) {
                const output = yield this.dockerContainer.getStats();
                this.emit('container-stats', {
                    notif: 'Container stats',
                    status: true,
                    log: output
                });
            }
            else {
                this.emit('container-stats', {
                    status: false,
                    notif: 'Container not running'
                });
            }
        });
    }
    /**
     * Gets docker container logs then sends the result through socket
     */
    dockerLogs() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.dockerContainer) {
                this.socket.emit('dockerLogs', yield this.dockerContainer.getLogs(this.clogsSince));
            }
        });
    }
    /**
     * Starts the container with the povided data
     *
     * @param  {IStartContainer} data Image name and envs
     * @return {Promise}
     */
    startContainer(data) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Cli_1.default.checkDocker();
            if (data && data.image.length > 0) {
                try {
                    !utils.doesContainerExist(data.image, yield utils.docker.listImages());
                }
                catch (e) {
                    this.emit('start', {
                        notif: `Image for microservice is not built. Run \`omg build\` to build the image.`,
                        status: false
                    });
                    return;
                }
            }
            else {
                yield this.buildImage({});
            }
            let envObj;
            try {
                envObj = data.envs;
                envObj = utils.matchEnvironmentCases(envObj, this.microservice.environmentVariables);
            }
            catch (e) {
                this.emit('start', {
                    status: false,
                    notif: e
                });
                return;
            }
            if (this.rebuildBak === undefined) {
                this.rebuildBak = { start: data, build: {} };
            }
            this.dockerContainer = new RunFactory_1.default(data.image, this.microservice, null, envObj).getRun(null, true);
            // Start container
            this.socket.emit('start', {
                notif: 'Starting Docker container',
                status: true
            });
            this.containerID = yield this.dockerContainer.startService(this.inheritEnv);
            this.socket.emit('start', {
                notif: `Started Docker container: ${this.containerID.substring(0, 12)}`,
                status: true,
                started: true,
                ports: this.dockerContainer.getPortbindings(),
                forwards: this.dockerContainer.forwardPortsBindings
            });
            yield new Promise(res => setTimeout(res, 1000));
        });
    }
    /**
     * Stops the currently running container
     *
     * @return {Promise}
     */
    stopContainer() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            let output;
            if (this.dockerContainer !== null &&
                (yield this.dockerContainer.isRunning())) {
                output = yield this.dockerContainer.stopService();
                this.emit('stop', {
                    status: true,
                    notif: `Stoppped Docker container: ${output}`
                });
            }
            resolve();
        }));
    }
    /**
     * Removes the current container
     *
     * @return {Promise}
     */
    removeContainer() {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            let output;
            if (this.dockerContainer !== null &&
                !(yield this.dockerContainer.isRunning())) {
                output = yield this.dockerContainer.removeContainer();
                this.emit('stop', {
                    status: true,
                    notif: `Removed Docker container: ${output}`
                });
            }
            this.dockerContainer = null;
            resolve();
        }));
    }
    /**
     * Run the asked action with params
     *
     * @param  {any} data actions and its params
     */
    runAction(data) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.healthCheck();
            // Run action
            this.socket.emit('run', {
                notif: `Running action: \`${data.action}\``,
                status: true
            });
            let output;
            try {
                this.dockerContainer.setArgs(data.args);
                output = yield this.dockerContainer.exec(data.action);
                this.socket.emit('run', {
                    notif: `Ran action: \`${data.action}\` with output: ${output}`,
                    status: true
                });
            }
            catch (e) {
                this.socket.emit('run', {
                    notif: `Failed action: \`${data.action}\`: ${e}`,
                    output: e.error,
                    done: true,
                    status: false
                });
                return;
            }
            this.socket.emit('run', {
                output: output,
                notif: output,
                done: true,
                status: true
            });
        });
    }
    /**
     * Subscribes to the asked event eith provided params
     *
     * @param  {IDataAction} data
     */
    subscribeEvent(data) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Cli_1.default.checkDocker();
            yield this.runAction({
                action: data.action,
                args: data.args,
                image: data.image
            });
            this.socket.emit('subscribe', {
                notif: `Subscribing to event: \`${data.event}\``,
                status: true
            });
            this.subscribe = new Subscribe_1.default(this.microservice, data.args, this.dockerContainer);
            try {
                yield this.subscribe.go(data.action, data.event);
                this.socket.emit('subscribe', {
                    notif: `Subscribed to event: \`${data.event}\` data will be posted to this terminal window when appropriate`,
                    status: true
                });
                setInterval(() => __awaiter(this, void 0, void 0, function* () {
                    if (!this.dockerContainer) {
                        return;
                    }
                    if (!(yield this.dockerContainer.isRunning())) {
                        this.socket.emit('subscribe', {
                            notif: 'Container unexpectedly stopped',
                            status: false,
                            logs: `${yield this.dockerContainer.getStderr()}`
                        });
                        return;
                    }
                }), 1500);
            }
            catch (e) {
                this.socket.emit('subscribe', {
                    notif: `Failed subscribing to event ${data.event}: ${e}`,
                    status: false,
                    logs: `${yield this.dockerContainer.getStderr()}`
                });
                return;
            }
        });
    }
    /**
     * Gets docker container health status then sends the result through socket
     */
    healthCheck() {
        return __awaiter(this, void 0, void 0, function* () {
            let isHealthy;
            try {
                isHealthy = yield this.dockerContainer.healthCheck();
            }
            catch (e) {
                isHealthy = false;
            }
            if (!isHealthy) {
                try {
                    if (this.dockerContainer !== null &&
                        (yield this.dockerContainer.isRunning())) {
                        this.socket.emit('health-check', {
                            notif: 'Health check failed',
                            status: 0
                        });
                    }
                    else {
                        this.socket.emit('health-check', {
                            notif: 'Health check failed',
                            status: -1,
                            log: 'Unable to retrieve error logs'
                        });
                    }
                }
                catch (_a) {
                    this.socket.emit('health-check', {
                        notif: 'Health check failed',
                        status: -1,
                        log: 'Unable to retrieve error logs'
                    });
                }
            }
            else {
                this.socket.emit('health-check', {
                    notif: 'Health check passed',
                    status: 1
                });
            }
        });
    }
}
exports.default = UIServer;
//# sourceMappingURL=UI.js.map