"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const $ = require("shelljs");
const net = require("http");
const fs = require("fs");
const path = require("path");
const Docker = require('dockerode-promise');
const LineUp = require('lineup');
const execCmd = require('child_process').exec;
const lineup = new LineUp();
exports.docker = new Docker();
let versionAvailable = false;
/**
 * Used to set values in the constructors of the microservice classes.
 *
 * @param {*} val The value to set
 * @param {*} _else The value to set if val if not defined
 * @return {*} The value
 */
function setVal(val, _else) {
    if (_.isUndefined(val)) {
        return _else;
    }
    return val;
}
exports.setVal = setVal;
/**
 * @param  {number} ms Time to sleep in milliseconds
 * @return {Promise} Promise to await
 */
function sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}
exports.sleep = sleep;
/**
 * Get's the ports that need to be open defined by the given {@link Microservice}.
 *
 * @param {Microservice} microservice The given {@link Microservice}
 * @return {Array<Integer>} The ports that need to be opened for the given {@link Microservice}
 */
function getNeededPorts(microservice) {
    const ports = [];
    for (let i = 0; i < microservice.actions.length; i += 1) {
        const action = microservice.actions[i];
        if (action.http !== null) {
            if (!ports.includes(action.http.port)) {
                ports.push(action.http.port);
            }
        }
        if (action.events !== null) {
            for (let j = 0; j < action.events.length; j += 1) {
                if (!ports.includes(action.events[j].subscribe.port)) {
                    ports.push(action.events[j].subscribe.port);
                }
                if (action.events[j].unsubscribe &&
                    !ports.includes(action.events[j].unsubscribe.port)) {
                    ports.push(action.events[j].unsubscribe.port);
                }
            }
        }
    }
    for (let i = 0; i < microservice.forwards.length; i++) {
        const forward = microservice.forwards[i];
        if (forward.http !== null) {
            ports.push(forward.http.port);
        }
    }
    if (microservice.health &&
        microservice.health.port !== null &&
        !ports.includes(microservice.health.port)) {
        ports.push(microservice.health.port);
    }
    return ports;
}
exports.getNeededPorts = getNeededPorts;
/**
 * Gets the ports that need to be open defined by the given {@link Microservice}.
 *
 * @param  {Microservice} microservice
 * @return {number[]} ports Bond ports
 */
function getForwardPorts(microservice) {
    const ports = [];
    for (let i = 0; i < microservice.forwards.length; i++) {
        const forward = microservice.forwards[i];
        if (forward.http !== null) {
            ports.push(forward.http.port);
        }
    }
    return ports;
}
exports.getForwardPorts = getForwardPorts;
/**
 * Return the health port
 *
 * @param  {Microservice} microservice Provided microservice as a JSON
 * @return {number} port number
 */
function getHealthPort(microservice) {
    return microservice.health.port;
}
exports.getHealthPort = getHealthPort;
/**
 * Creates a name for the Docker images based of the git remote -v.
 *
 * @param {boolean} ui Defines if UI mode is enabled or not
 * @return {Promise<String>} The image name
 */
function createImageName(ui = false) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const data = yield exec('git remote -v');
            if (data.match(/git@github\.com:(\w+\/[\w|.|-]+).git/)) {
                return ui
                    ? `${data
                        .match(/git@github\.com:(\w+\/[\w|.|-]+).git/)[1]
                        .toLowerCase()}`
                    : `omg/${data
                        .match(/git@github\.com:(\w+\/[\w|.|-]+).git/)[1]
                        .toLowerCase()}`;
            }
            else {
                return ui
                    ? `${data
                        .match(/https:\/\/github\.com\/(\w+\/[\w|.|-]+)/)[1]
                        .toLowerCase()}`
                    : `omg/${data
                        .match(/https:\/\/github\.com\/(\w+\/[\w|.|-]+)/)[1]
                        .toLowerCase()}`;
            }
        }
        catch (e) {
            if (ui) {
                return {
                    owner: `${Buffer.from(process.cwd())
                        .toString('base64')
                        .toLowerCase()
                        .replace(/=/g, '')}`,
                    generated: true
                };
            }
            else {
                return `omg/${Buffer.from(process.cwd())
                    .toString('base64')
                    .toLowerCase()
                    .replace(/=/g, '')}`;
            }
        }
    });
}
exports.createImageName = createImageName;
/**
 * Turns a list of string with a delimiter to a map.
 *
 * @param {Array<String>} list The given list of strings with delimiter
 * @param {String} errorMessage The given message to used when unable to parse
 * @return {Object} Key value of the list
 */
function parse(list, errorMessage) {
    const dictionary = {};
    for (let i = 0; i < list.length; i += 1) {
        const split = list[i].split(/=(.+)/);
        if (split.length !== 3) {
            throw {
                message: errorMessage
            };
        }
        dictionary[split[0]] = split[1];
    }
    return dictionary;
}
exports.parse = parse;
/**
 * Matches the case of given cli environment arguments to the case defined in
 * the microservice.yml.
 *
 * @param {Object} env The given environment variable mapping
 * @param {Array<EnvironmentVariable>} environmentVariables The given {@link EnvironmentVariable}s
 * @return {Object} The environment mapping with correct cases
 */
function matchEnvironmentCases(env, environmentVariables) {
    const result = {};
    for (const key of Object.keys(env)) {
        for (const cap of environmentVariables) {
            if (cap.name.toLowerCase() === key.toLowerCase()) {
                result[cap.name] = env[key];
            }
        }
    }
    return result;
}
exports.matchEnvironmentCases = matchEnvironmentCases;
/**
 * Promise wrapper for the `exec`.
 *
 * @param {String} command The command to run
 * @param {Boolean} [silent=true] True if silent, otherwise false
 * @return {Promise<String>} The stdout if resolved, otherwise stderror unless stderror is empty
 */
function exec(command, silent = true) {
    return new Promise(function (resolve, reject) {
        $.exec(command, { silent }, function (code, stdout, stderr) {
            if (code !== 0) {
                if (stderr === '') {
                    reject(stdout.trim());
                }
                else {
                    reject(stderr.trim());
                }
            }
            else {
                if (stdout === '') {
                    resolve(stderr.trim());
                }
                else {
                    resolve(stdout.trim());
                }
            }
        });
    });
}
exports.exec = exec;
/**
 * Checks if a docker container exists with the given name.
 *
 * @param {String} name The given name
 * @param {Array<Object>} containers The given containers
 * @return {Boolean} True if it exists, otherwise, false
 */
function doesContainerExist(name, containers) {
    for (let i = 0; i < containers.length; i += 1) {
        for (let j = 0; j < containers[i].RepoTags.length; j += 1) {
            if (containers[i].RepoTags[j].includes(name)) {
                return true;
            }
        }
    }
    return false;
}
exports.doesContainerExist = doesContainerExist;
/**
 * Log a string to stdout.
 *
 * @param {String} string The given sting to log
 */
function log(string) {
    process.stdout.write(`${string}\n`);
}
exports.log = log;
/**
 * Log a string to stderr.
 *
 * @param {String} string The given string to log
 */
function error(string) {
    process.stderr.write(`${string}\n`);
}
exports.error = error;
exports.typeCast = {
    int: (int) => parseInt(int),
    float: (float) => parseFloat(float),
    string: (string) => string,
    uuid: (uuid) => uuid,
    list: (list) => JSON.parse(list),
    map: (map) => JSON.parse(map),
    object: (object) => JSON.parse(object),
    boolean: (boolean) => boolean === 'true',
    path: (path) => path,
    any: (any) => any
};
exports.dataTypes = {
    int: (int) => {
        return int.toString().match(/^[-+]?\d+$/) !== null;
    },
    float: (float) => {
        return (!isNaN(parseFloat(float)) &&
            parseFloat(float)
                .toString()
                .indexOf('.') !== -1);
    },
    string: (string) => {
        return true;
    },
    uuid: (uuid) => {
        return (uuid.match(/^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$/) !== null);
    },
    list: (list) => {
        try {
            return ((Array.isArray(list) && typeof list === 'object') ||
                JSON.parse(list).toString() !== '[object Object]');
        }
        catch (e) {
            return false;
        }
    },
    map: (map) => {
        try {
            return ((!Array.isArray(map) && typeof map === 'object') ||
                JSON.parse(map).toString() === '[object Object]');
        }
        catch (e) {
            return false;
        }
    },
    object: (object) => {
        try {
            return ((!Array.isArray(object) && typeof object === 'object') ||
                JSON.parse(object).toString() === '[object Object]');
        }
        catch (e) {
            return false;
        }
    },
    boolean: (boolean) => {
        return boolean === 'false' || boolean === 'true';
    },
    path: (path) => {
        try {
            JSON.parse(path);
            return false;
        }
        catch (e) {
            const lastChar = path.substr(path.length - 1);
            if (lastChar === '/') {
                return false;
            }
            return typeof path === 'string';
        }
    },
    any: (any) => {
        return true;
    }
};
/**
 * Get's an open port.
 *
 * @return {Promise<Number>} The open port
 */
function getOpenPort() {
    return new Promise(resolve => {
        getPort(data => {
            resolve(data);
        });
    });
    /**
     * Helper for {@link getOpenPort}. Tries to open a server on a random port, if fail try again, otherwise call
     * the callback with the port.
     *
     * @param {Function} cb The given callback
     */
    function getPort(cb) {
        const server = net.createServer();
        const port = Math.floor(Math.random() * 15000) + 2000; // port range 2000 to 17000
        server.listen(port, () => {
            server.once('close', () => {
                cb(port);
            });
            server.close();
        });
        server.on('error', () => {
            getPort(cb);
        });
    }
}
exports.getOpenPort = getOpenPort;
/**
 * Used to append the environment variable options into [args...].
 *
 * @param {Array} xs
 * @return {function(*=): (*|Array)}
 */
function appender(xs) {
    xs = xs || [];
    return function (x) {
        xs.push(x);
        return xs;
    };
}
exports.appender = appender;
/**
 * Checks if any action in he given microserviceJson does not have
 * and interface (http, format, rpc, xor events).
 *
 * @param {Object} microserviceJson The given microservice json
 * @throws {Object} Throws error if more, or less than one interface
 *                  is given for an action
 */
function checkActionInterface(microserviceJson) {
    if (microserviceJson.actions) {
        const actionMap = microserviceJson.actions;
        for (const actionName of Object.keys(actionMap)) {
            const action = actionMap[actionName];
            const bools = [
                !!action.http,
                !!action.format,
                !!action.rpc,
                !!action.events
            ].filter(b => b);
            if (bools.length !== 1) {
                throw {
                    text: `actions.${actionName} should have one of required property: 'http' 'format' 'rpc' or 'events'`
                };
            }
        }
    }
}
exports.checkActionInterface = checkActionInterface;
/**
 * Checks if the current version is up to date with the published one
 */
function checkVersion() {
    execCmd('npm view omg version', {
        encoding: 'utf8'
    }, (e, out, err) => {
        if (out) {
            const versions = {
                local: require('../package.json')
                    .version.trim()
                    .match(/^(\d+).(\d+).(\d+)/),
                distant: out
                    .toString()
                    .trim()
                    .match(/^(\d+).(\d+).(\d+)/)
            };
            for (let i = 1; i <= 3; i++) {
                if (versions.distant[i] > versions.local[i]) {
                    lineup.sticker.note('');
                    lineup.sticker.note(`${lineup.colors.yellow(`${i === 1 ? 'Major' : i === 2 ? 'Minor' : 'Patch'} update available: `)}${lineup.colors.red(versions.local[0])} ${lineup.colors.yellow('=>')} ${lineup.colors.green(versions.distant[0])}`);
                    lineup.sticker.note(`${lineup.colors.yellow(`Run: 'npm i -g omg' or 'yarn global add omg' to update`)}`);
                    lineup.sticker.note('');
                    versionAvailable = true;
                }
            }
        }
    });
}
exports.checkVersion = checkVersion;
/**
 * If the CLI is outdated, posts a warning in the terminal
 */
function showVersionCard() {
    if (versionAvailable) {
        lineup.sticker.show({
            align: 'center',
            color: 'red'
        });
    }
}
exports.showVersionCard = showVersionCard;
/**
 * @return {string} Content of the microservice file
 */
function readMicroserviceFile() {
    let content = '';
    try {
        content = fs
            .readFileSync(path.join(process.cwd(), 'microservice.yml'))
            .toString();
    }
    catch (_a) {
        try {
            content = fs
                .readFileSync(path.join(process.cwd(), 'microservice.yaml'))
                .toString();
        }
        catch (_b) {
            content = '';
        }
    }
    return content;
}
exports.readMicroserviceFile = readMicroserviceFile;
/**
 * @return {string} microservice file path
 */
function getMicroserviceFilePath() {
    if (fs.existsSync(path.join(process.cwd(), 'microservice.yml'))) {
        return path.join(process.cwd(), 'microservice.yml');
    }
    else if (fs.existsSync(path.join(process.cwd(), 'microservice.yaml'))) {
        return path.join(process.cwd(), 'microservice.yaml');
    }
    else {
        return null;
    }
}
exports.getMicroserviceFilePath = getMicroserviceFilePath;
//# sourceMappingURL=utils.js.map