"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const utils_1 = require("./utils");
/**
 * Verifies the pattern, enum, and range of the given arguments.
 *
 * @param {Command} command The given {@link Command}
 * @param {Object} args The given argument mapping
 */
function verifyArgumentConstrains(command, args) {
    command.arguments.forEach(a => {
        if (Object.keys(args).includes(a.name)) {
            const argumentValue = args[a.name];
            if (a.pattern !== null && argumentValue.match(a.pattern) === null) {
                throw `Argument: \`${a.name}\` must match regex: \`${a.pattern}\``;
            }
            if (a.enum !== null && !a.enum.includes(argumentValue)) {
                throw `Argument: \`${a.name}\` must be one of: \`${a.enum.toString()}\``;
            }
            if (a.range !== null) {
                if (!_.isUndefined(a.range.min) && argumentValue < a.range.min) {
                    throw `Argument: \`${a.name}\` must be be no smaller than the value: \`${a.range.min}\``;
                }
                if (!_.isUndefined(a.range.max) && argumentValue > a.range.max) {
                    throw `Argument: \`${a.name}\` must be be no larger than the value: \`${a.range.max}\``;
                }
            }
        }
    });
}
exports.verifyArgumentConstrains = verifyArgumentConstrains;
/**
 * Verifies the types of given arguments based off of the given {@link Action} or {@link Event}.
 *
 * @param {Command} command The given {@link Action} or {@link Event}
 * @param {Object} args The given argument mapping
 */
function verifyArgumentTypes(command, args) {
    command.arguments.forEach(a => {
        if (Object.keys(args).includes(a.name)) {
            if (!utils_1.dataTypes[a.type](args[a.name])) {
                throw `Argument: \`${a.name}\` must be of type: \`${a.type}\``;
            }
        }
    });
}
exports.verifyArgumentTypes = verifyArgumentTypes;
/**
 * Verifies the types of the environment variables.
 *
 * @param {Microservice} microservice The given {@link Microservice}
 * @param {Object} envs The given environment variable mapping
 */
function verifyEnvironmentVariableTypes(microservice, envs) {
    microservice.environmentVariables.forEach(e => {
        if (Object.keys(envs).includes(e.name)) {
            if (!utils_1.dataTypes[e.type](envs[e.name])) {
                throw `Environment variable: \`${e.name}\` must be of type: \`${e.type}\``;
            }
        }
    });
}
exports.verifyEnvironmentVariableTypes = verifyEnvironmentVariableTypes;
/**
 * Verifies the patterns of the environment variables, if a pattern is defined.
 *
 * @param {Microservice} microservice The given {@link Microservice}
 * @param {Object} envs The given environment variable mapping
 */
function verifyEnvironmentVariablePattern(microservice, envs) {
    microservice.environmentVariables.forEach(e => {
        if (Object.keys(envs).includes(e.name)) {
            if (e.pattern !== null && envs[e.name].match(e.pattern) === null) {
                throw `Environment variable: \`${e.name}\` must match regex: \`${e.pattern}\``;
            }
        }
    });
}
exports.verifyEnvironmentVariablePattern = verifyEnvironmentVariablePattern;
/**
 * Verifies the output type of a container.
 *
 * @param {Command} command The given {@link Action}
 * @param {String} output The given output
 */
function verifyOutputType(command, output) {
    if (command.output &&
        command.output.type &&
        !utils_1.dataTypes[command.output.type](output)) {
        throw `Action: \`${command.name}\`` +
            ` must have output type: \`${command.output.type}\`` +
            ` instead got: \`${typeof output}\`` +
            ` ${output}`;
    }
    else if (command.output &&
        command.output.type &&
        command.output.type === 'object') {
        const props = command.output.properties;
        const json = JSON.parse(output);
        Object.keys(props).forEach(key => {
            if (props[key].type in utils_1.dataTypes &&
                !utils_1.dataTypes[props[key].type](JSON.stringify(json[key]))) {
                const typeFound = (json, key) => {
                    if (typeof json[key] === 'number' &&
                        utils_1.dataTypes['int'](json[key].toString())) {
                        return 'int';
                    }
                    else if (utils_1.dataTypes['float'](json[key].toString())) {
                        return 'float';
                    }
                    return typeof json[key];
                };
                throw `Action: \`${command.name}\`` +
                    ` property \`${key}\` must have output type: \`${props[key].type}\`` +
                    ` instead got: \`${typeFound(json, key)}\`` +
                    `\n${output}`;
            }
        });
    }
}
exports.verifyOutputType = verifyOutputType;
/**
 * If defined, verifies the output properties of a map or object.
 *
 * @param {Command} command  The given {@link Command}
 * @param {String} output The given output
 */
function verifyProperties(command, output) {
    const outputObject = JSON.parse(output);
    const properties = command.output.properties;
    const outputedProperties = Object.keys(outputObject);
    for (const property of Object.keys(properties)) {
        if (!(property in outputObject)) {
            throw `Action: \`${command.name}\`'s output does not have expected property: \`${property}\``;
        }
        if (!utils_1.dataTypes[properties[property].type](JSON.stringify(outputObject[property]))) {
            throw `Action: \`${command.name}\`'s output property: \`${property}\`` +
                ` must have type: \`${properties[property].type}\`` +
                ` instead got: \`${typeof outputObject[property]}\`:` +
                ` ${JSON.stringify(outputObject[property])}`;
        }
        outputedProperties.splice(outputedProperties.indexOf(property), 1);
    }
    const extraPropertyCount = outputedProperties.length;
    if (extraPropertyCount > 0) {
        throw `Action: \`${command.name}\` outputed ${extraPropertyCount} extra propert${extraPropertyCount === 1 ? 'y' : 'ies'}: \`${outputedProperties.toString()}\``;
    }
}
exports.verifyProperties = verifyProperties;
//# sourceMappingURL=verify.js.map